# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: caffe2/proto/caffe2.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x19\x63\x61\x66\x66\x65\x32/proto/caffe2.proto\x12\x06\x63\x61\x66\x66\x65\x32\"\xa3\x05\n\x0bTensorProto\x12\x0c\n\x04\x64ims\x18\x01 \x03(\x03\x12\x36\n\tdata_type\x18\x02 \x01(\x0e\x32\x1c.caffe2.TensorProto.DataType:\x05\x46LOAT\x12\x16\n\x0b\x64\x61ta_format\x18\x0f \x01(\r:\x01\x30\x12\x16\n\nfloat_data\x18\x03 \x03(\x02\x42\x02\x10\x01\x12\x16\n\nint32_data\x18\x04 \x03(\x05\x42\x02\x10\x01\x12\x11\n\tbyte_data\x18\x05 \x01(\x0c\x12\x13\n\x0bstring_data\x18\x06 \x03(\x0c\x12\x17\n\x0b\x64ouble_data\x18\t \x03(\x01\x42\x02\x10\x01\x12\x16\n\nint64_data\x18\n \x03(\x03\x42\x02\x10\x01\x12\x10\n\x08raw_data\x18\r \x01(\x0c\x12\x0c\n\x04name\x18\x07 \x01(\t\x12+\n\rdevice_detail\x18\x08 \x01(\x0b\x32\x14.caffe2.DeviceOption\x12,\n\x07segment\x18\x0b \x01(\x0b\x32\x1b.caffe2.TensorProto.Segment\x1a%\n\x07Segment\x12\r\n\x05\x62\x65gin\x18\x01 \x02(\x03\x12\x0b\n\x03\x65nd\x18\x02 \x02(\x03\"\xcf\x01\n\x08\x44\x61taType\x12\r\n\tUNDEFINED\x10\x00\x12\t\n\x05\x46LOAT\x10\x01\x12\t\n\x05INT32\x10\x02\x12\x08\n\x04\x42YTE\x10\x03\x12\n\n\x06STRING\x10\x04\x12\x08\n\x04\x42OOL\x10\x05\x12\t\n\x05UINT8\x10\x06\x12\x08\n\x04INT8\x10\x07\x12\n\n\x06UINT16\x10\x08\x12\t\n\x05INT16\x10\t\x12\t\n\x05INT64\x10\n\x12\x0b\n\x07\x46LOAT16\x10\x0c\x12\n\n\x06\x44OUBLE\x10\r\x12\x17\n\x13ZERO_COLLISION_HASH\x10\x0e\x12\x15\n\x11REBATCHING_BUFFER\x10\x0f\"9\n\x13SerializationFormat\x12\x10\n\x0c\x46MT_PROTOBUF\x10\x00\x12\x10\n\x0c\x46MT_BFLOAT16\x10\x01\"\x83\x02\n\x0cQTensorProto\x12\x0c\n\x04\x64ims\x18\x01 \x03(\x03\x12\x11\n\tprecision\x18\x02 \x02(\x05\x12\r\n\x05scale\x18\x03 \x02(\x01\x12\x0c\n\x04\x62ias\x18\x04 \x02(\x01\x12\x11\n\tis_signed\x18\x05 \x02(\x08\x12\x10\n\x04\x64\x61ta\x18\x06 \x03(\x05\x42\x02\x10\x01\x12\x0c\n\x04name\x18\x07 \x01(\t\x12\x36\n\tdata_type\x18\x08 \x01(\x0e\x32\x1c.caffe2.TensorProto.DataType:\x05INT32\x12\x0e\n\x06scales\x18\t \x03(\x01\x12\x0e\n\x06\x62iases\x18\n \x03(\x01\x12\x0c\n\x04\x61xis\x18\x0b \x01(\x05\x12\x1c\n\ris_multiparam\x18\x0c \x01(\x08:\x05\x66\x61lse\"3\n\x0cTensorProtos\x12#\n\x06protos\x18\x01 \x03(\x0b\x32\x13.caffe2.TensorProto\"\x95\x01\n\x0bTensorShape\x12\x0c\n\x04\x64ims\x18\x01 \x03(\x03\x12\x36\n\tdata_type\x18\x02 \x01(\x0e\x32\x1c.caffe2.TensorProto.DataType:\x05\x46LOAT\x12\x14\n\x0cunknown_dims\x18\x03 \x03(\x05\x12\x1c\n\runknown_shape\x18\x04 \x01(\x08:\x05\x66\x61lse\x12\x0c\n\x04name\x18\x05 \x01(\t\"3\n\x0cTensorShapes\x12#\n\x06shapes\x18\x01 \x03(\x0b\x32\x13.caffe2.TensorShape\"\xa8\x02\n\x10TensorBoundShape\x12\"\n\x05shape\x18\x01 \x01(\x0b\x32\x13.caffe2.TensorShape\x12\x32\n\x08\x64im_type\x18\x02 \x03(\x0e\x32 .caffe2.TensorBoundShape.DimType\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x16\n\x0eshape_is_final\x18\x04 \x01(\x08\"\x95\x01\n\x07\x44imType\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0c\n\x08\x43ONSTANT\x10\x01\x12\t\n\x05\x42\x41TCH\x10\x02\x12\x18\n\x14\x42\x41TCH_OF_FEATURE_MAX\x10\x03\x12 \n\x1c\x42\x41TCH_OF_FEATURE_MAX_DEFAULT\x10\x04\x12\x0f\n\x0b\x46\x45\x41TURE_MAX\x10\x05\x12\x17\n\x13\x46\x45\x41TURE_MAX_DEFAULT\x10\x06\"n\n\x11TensorBoundShapes\x12(\n\x06shapes\x18\x01 \x03(\x0b\x32\x18.caffe2.TensorBoundShape\x12\x16\n\x0emax_batch_size\x18\x02 \x01(\x03\x12\x17\n\x0fmax_feature_len\x18\x03 \x01(\x03\"\x8d\x01\n\tAOTConfig\x12\x16\n\x0emax_batch_size\x18\x01 \x02(\x03\x12\x14\n\x0cmax_seq_size\x18\x02 \x02(\x03\x12\x1a\n\x12in_batch_broadcast\x18\x03 \x02(\x08\x12\x1d\n\x15onnxifi_blacklist_ops\x18\x04 \x01(\t\x12\x17\n\x0fonnxifi_min_ops\x18\x05 \x01(\x05\"\x8f\x02\n\x08\x41rgument\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\t\n\x01\x66\x18\x02 \x01(\x02\x12\t\n\x01i\x18\x03 \x01(\x03\x12\t\n\x01s\x18\x04 \x01(\x0c\x12\x1e\n\x01t\x18\n \x01(\x0b\x32\x13.caffe2.TensorProto\x12\x19\n\x01n\x18\x08 \x01(\x0b\x32\x0e.caffe2.NetDef\x12\x0e\n\x06\x66loats\x18\x05 \x03(\x02\x12\x0c\n\x04ints\x18\x06 \x03(\x03\x12\x0f\n\x07strings\x18\x07 \x03(\x0c\x12$\n\x07tensors\x18\x0b \x03(\x0b\x32\x13.caffe2.TensorProto\x12\x1c\n\x04nets\x18\t \x03(\x0b\x32\x0e.caffe2.NetDef\x12&\n\x08qtensors\x18\x0c \x03(\x0b\x32\x14.caffe2.QTensorProto\"\x8b\x01\n\x0c\x44\x65viceOption\x12\x16\n\x0b\x64\x65vice_type\x18\x01 \x01(\x05:\x01\x30\x12\x11\n\tdevice_id\x18\x02 \x01(\x05\x12\x13\n\x0brandom_seed\x18\x03 \x01(\r\x12\x11\n\tnode_name\x18\x04 \x01(\t\x12\x14\n\x0cnuma_node_id\x18\x05 \x01(\x05\x12\x12\n\nextra_info\x18\x06 \x03(\t\"\x92\x02\n\x0bOperatorDef\x12\r\n\x05input\x18\x01 \x03(\t\x12\x0e\n\x06output\x18\x02 \x03(\t\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x0c\n\x04type\x18\x04 \x01(\t\x12\x1d\n\x03\x61rg\x18\x05 \x03(\x0b\x32\x10.caffe2.Argument\x12+\n\rdevice_option\x18\x06 \x01(\x0b\x32\x14.caffe2.DeviceOption\x12\x0e\n\x06\x65ngine\x18\x07 \x01(\t\x12\x15\n\rcontrol_input\x18\x08 \x03(\t\x12\x1d\n\x0eis_gradient_op\x18\t \x01(\x08:\x05\x66\x61lse\x12\x12\n\ndebug_info\x18\n \x01(\t\x12\x0e\n\x06\x64omain\x18\x0b \x01(\t\x12\x12\n\nop_version\x18\x0c \x01(\x03\")\n\rMapFieldEntry\x12\x0b\n\x03key\x18\x01 \x02(\t\x12\x0b\n\x03val\x18\x02 \x02(\t\"M\n\x0e\x42\x61\x63kendOptions\x12\x14\n\x0c\x62\x61\x63kend_name\x18\x01 \x02(\t\x12%\n\x06option\x18\x02 \x03(\x0b\x32\x15.caffe2.MapFieldEntry\"u\n\rPartitionInfo\x12\x0c\n\x04name\x18\x01 \x02(\t\x12\x11\n\tdevice_id\x18\x02 \x03(\x05\x12\x12\n\nextra_info\x18\x03 \x01(\t\x12/\n\x0f\x62\x61\x63kend_options\x18\x04 \x03(\x0b\x32\x16.caffe2.BackendOptions\"\x86\x02\n\x06NetDef\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1f\n\x02op\x18\x02 \x03(\x0b\x32\x13.caffe2.OperatorDef\x12\x0c\n\x04type\x18\x03 \x01(\t\x12\x13\n\x0bnum_workers\x18\x04 \x01(\x05\x12+\n\rdevice_option\x18\x05 \x01(\x0b\x32\x14.caffe2.DeviceOption\x12\x1d\n\x03\x61rg\x18\x06 \x03(\x0b\x32\x10.caffe2.Argument\x12\x16\n\x0e\x65xternal_input\x18\x07 \x03(\t\x12\x17\n\x0f\x65xternal_output\x18\x08 \x03(\t\x12-\n\x0epartition_info\x18\t \x03(\x0b\x32\x15.caffe2.PartitionInfo\"\xcf\x02\n\rExecutionStep\x12\x0c\n\x04name\x18\x01 \x01(\t\x12&\n\x07substep\x18\x02 \x03(\x0b\x32\x15.caffe2.ExecutionStep\x12\x0f\n\x07network\x18\x03 \x03(\t\x12\x10\n\x08num_iter\x18\x04 \x01(\x03\x12\x1c\n\x10\x63riteria_network\x18\x05 \x01(\tB\x02\x18\x01\x12\x12\n\nreport_net\x18\x07 \x01(\t\x12\x17\n\x0freport_interval\x18\x08 \x01(\x05\x12\x14\n\x0crun_every_ms\x18\x0b \x01(\x03\x12\x1b\n\x13\x63oncurrent_substeps\x18\x06 \x01(\x08\x12\x18\n\x10should_stop_blob\x18\t \x01(\t\x12\x11\n\tonly_once\x18\n \x01(\x08\x12\x18\n\x10\x63reate_workspace\x18\x0c \x01(\x08\x12 \n\x18num_concurrent_instances\x18\r \x01(\x05\"g\n\x07PlanDef\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1f\n\x07network\x18\x02 \x03(\x0b\x32\x0e.caffe2.NetDef\x12-\n\x0e\x65xecution_step\x18\x03 \x03(\x0b\x32\x15.caffe2.ExecutionStep\"\xba\x01\n\tBlobProto\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04type\x18\x02 \x01(\t\x12#\n\x06tensor\x18\x03 \x01(\x0b\x32\x13.caffe2.TensorProto\x12\x0f\n\x07\x63ontent\x18\x04 \x01(\x0c\x12%\n\x07qtensor\x18\x05 \x01(\x0b\x32\x14.caffe2.QTensorProto\x12\x1a\n\x12\x63ontent_num_chunks\x18\x06 \x01(\x05\x12\x18\n\x10\x63ontent_chunk_id\x18\x07 \x01(\x05\"K\n\rDBReaderProto\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0e\n\x06source\x18\x02 \x01(\t\x12\x0f\n\x07\x64\x62_type\x18\x03 \x01(\t\x12\x0b\n\x03key\x18\x04 \x01(\t\"\xd5\x01\n\x18\x42lobSerializationOptions\x12\x17\n\x0f\x62lob_name_regex\x18\x01 \x01(\t\x12\x12\n\nchunk_size\x18\x02 \x01(\x03\x12\x42\n\x0c\x66loat_format\x18\x03 \x01(\x0e\x32,.caffe2.BlobSerializationOptions.FloatFormat\"H\n\x0b\x46loatFormat\x12\x11\n\rFLOAT_DEFAULT\x10\x00\x12\x12\n\x0e\x46LOAT_PROTOBUF\x10\x01\x12\x12\n\x0e\x46LOAT_BFLOAT16\x10\x02\"I\n\x14SerializationOptions\x12\x31\n\x07options\x18\x01 \x03(\x0b\x32 .caffe2.BlobSerializationOptions*\xec\x01\n\x0f\x44\x65viceTypeProto\x12\r\n\tPROTO_CPU\x10\x00\x12\x0e\n\nPROTO_CUDA\x10\x01\x12\x10\n\x0cPROTO_MKLDNN\x10\x02\x12\x10\n\x0cPROTO_OPENGL\x10\x03\x12\x10\n\x0cPROTO_OPENCL\x10\x04\x12\x0f\n\x0bPROTO_IDEEP\x10\x05\x12\r\n\tPROTO_HIP\x10\x06\x12\x0e\n\nPROTO_FPGA\x10\x07\x12\r\n\tPROTO_ORT\x10\x08\x12\r\n\tPROTO_XLA\x10\t\x12\r\n\tPROTO_MPS\x10\n\x12\'\n#PROTO_COMPILE_TIME_MAX_DEVICE_TYPES\x10\x0b')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'caffe2.proto.caffe2_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _TENSORPROTO.fields_by_name['float_data']._options = None
  _TENSORPROTO.fields_by_name['float_data']._serialized_options = b'\020\001'
  _TENSORPROTO.fields_by_name['int32_data']._options = None
  _TENSORPROTO.fields_by_name['int32_data']._serialized_options = b'\020\001'
  _TENSORPROTO.fields_by_name['double_data']._options = None
  _TENSORPROTO.fields_by_name['double_data']._serialized_options = b'\020\001'
  _TENSORPROTO.fields_by_name['int64_data']._options = None
  _TENSORPROTO.fields_by_name['int64_data']._serialized_options = b'\020\001'
  _QTENSORPROTO.fields_by_name['data']._options = None
  _QTENSORPROTO.fields_by_name['data']._serialized_options = b'\020\001'
  _EXECUTIONSTEP.fields_by_name['criteria_network']._options = None
  _EXECUTIONSTEP.fields_by_name['criteria_network']._serialized_options = b'\030\001'
  _DEVICETYPEPROTO._serialized_start=3990
  _DEVICETYPEPROTO._serialized_end=4226
  _TENSORPROTO._serialized_start=38
  _TENSORPROTO._serialized_end=713
  _TENSORPROTO_SEGMENT._serialized_start=407
  _TENSORPROTO_SEGMENT._serialized_end=444
  _TENSORPROTO_DATATYPE._serialized_start=447
  _TENSORPROTO_DATATYPE._serialized_end=654
  _TENSORPROTO_SERIALIZATIONFORMAT._serialized_start=656
  _TENSORPROTO_SERIALIZATIONFORMAT._serialized_end=713
  _QTENSORPROTO._serialized_start=716
  _QTENSORPROTO._serialized_end=975
  _TENSORPROTOS._serialized_start=977
  _TENSORPROTOS._serialized_end=1028
  _TENSORSHAPE._serialized_start=1031
  _TENSORSHAPE._serialized_end=1180
  _TENSORSHAPES._serialized_start=1182
  _TENSORSHAPES._serialized_end=1233
  _TENSORBOUNDSHAPE._serialized_start=1236
  _TENSORBOUNDSHAPE._serialized_end=1532
  _TENSORBOUNDSHAPE_DIMTYPE._serialized_start=1383
  _TENSORBOUNDSHAPE_DIMTYPE._serialized_end=1532
  _TENSORBOUNDSHAPES._serialized_start=1534
  _TENSORBOUNDSHAPES._serialized_end=1644
  _AOTCONFIG._serialized_start=1647
  _AOTCONFIG._serialized_end=1788
  _ARGUMENT._serialized_start=1791
  _ARGUMENT._serialized_end=2062
  _DEVICEOPTION._serialized_start=2065
  _DEVICEOPTION._serialized_end=2204
  _OPERATORDEF._serialized_start=2207
  _OPERATORDEF._serialized_end=2481
  _MAPFIELDENTRY._serialized_start=2483
  _MAPFIELDENTRY._serialized_end=2524
  _BACKENDOPTIONS._serialized_start=2526
  _BACKENDOPTIONS._serialized_end=2603
  _PARTITIONINFO._serialized_start=2605
  _PARTITIONINFO._serialized_end=2722
  _NETDEF._serialized_start=2725
  _NETDEF._serialized_end=2987
  _EXECUTIONSTEP._serialized_start=2990
  _EXECUTIONSTEP._serialized_end=3325
  _PLANDEF._serialized_start=3327
  _PLANDEF._serialized_end=3430
  _BLOBPROTO._serialized_start=3433
  _BLOBPROTO._serialized_end=3619
  _DBREADERPROTO._serialized_start=3621
  _DBREADERPROTO._serialized_end=3696
  _BLOBSERIALIZATIONOPTIONS._serialized_start=3699
  _BLOBSERIALIZATIONOPTIONS._serialized_end=3912
  _BLOBSERIALIZATIONOPTIONS_FLOATFORMAT._serialized_start=3840
  _BLOBSERIALIZATIONOPTIONS_FLOATFORMAT._serialized_end=3912
  _SERIALIZATIONOPTIONS._serialized_start=3914
  _SERIALIZATIONOPTIONS._serialized_end=3987
# @@protoc_insertion_point(module_scope)
